from collections import namedtuple
from contextlib import contextmanager
from datetime import timedelta
from functools import wraps
from typing import List

from django.db import close_old_connections
from django.utils import timezone
from huey import Huey as HueyOriginal
from huey.api import Task
from huey.storage import RedisStorage, PriorityRedisStorage, RedisExpireStorage, PriorityRedisExpireStorage


class BaseHueyx(HueyOriginal):
    """
    Extends the Huey with new decorators which start a new transaction for every task.
    """

    DeadTask = namedtuple('DeadTask', ['id', 'name', 'settings'])
    HEARTBEAT_UPDATE_INTERVAL = 60  # min wait time in seconds to send another heartbeat to redis

    def db_task(self, *args, **kwargs):
        def decorator(fn):
            heartbeat_timeout = kwargs.pop('heartbeat_timeout', 0)
            if heartbeat_timeout:
                assert not kwargs.get('include_task',
                                      False), 'include_task and heartbeat_timeout keywords are not allowed together.'
                wrap = _wrap_heartbeat(close_db(fn, self), self, heartbeat_timeout)
                kwargs['context'] = True
            else:
                wrap = close_db(fn, self)
            task = self.task(*args, **kwargs)
            ret = task(wrap)
            ret.call_local = fn
            return ret

        return decorator

    def db_periodic_task(self, *args, **kwargs):
        def decorator(fn):
            return self.periodic_task(*args, **kwargs)(close_db(fn, self))

        return decorator

    @staticmethod
    def get_heartbeat_observation_key(task_id):
        return f'hb:{task_id}'

    @staticmethod
    def get_heartbeat_timestamp_key(task_id):
        return f'hbts:{task_id}'

    def get_dead_tasks(self) -> List[DeadTask]:
        dead_tasks = []
        observation_key_prefix = self.get_heartbeat_observation_key('')
        for result in self.storage.conn.hscan_iter(self.storage.result_key, match=observation_key_prefix + '*'):
            key = result[0].decode()
            task_id = key[len(observation_key_prefix):]
            name, task_settings, heartbeat_timeout = self.get(key, peek=True)
            timestamp = self.get(self.get_heartbeat_timestamp_key(task_id), peek=True)
            if not timestamp or timestamp + timedelta(seconds=heartbeat_timeout) <= timezone.now():
                dead_tasks.append(self.DeadTask(task_id, name, task_settings))
        return dead_tasks

    def restart_dead_tasks(self):
        for task in self.get_dead_tasks():
            task_type = self._registry.string_to_task(task.name)
            self.revoke_by_id(task.id)
            self.get(self.get_heartbeat_observation_key(task.id))
            task = task_type(**task.settings)
            self.enqueue(task)


class RedisHuey(BaseHueyx):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, storage_class=RedisStorage, **kwargs)


class PriorityRedisHuey(BaseHueyx):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, storage_class=PriorityRedisStorage, **kwargs)


class RedisExpireHuey(BaseHueyx):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, storage_class=RedisExpireStorage, **kwargs)


class PriorityRedisExpireHuey(BaseHueyx):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, storage_class=PriorityRedisExpireStorage, **kwargs)


def close_db(fn, huey: BaseHueyx):
    """Decorator to be used with tasks that may operate on the database."""

    @wraps(fn)
    def inner(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        finally:
            if not huey.immediate:
                close_old_connections()

    return inner


def _wrap_heartbeat(fn, huey: BaseHueyx, heartbeat_timeout: int):
    # noinspection PyProtectedMember
    @wraps(fn)
    def inner(*args, **kwargs):
        task: Task = kwargs.pop('task')
        heartbeat = Heartbeat(huey, task, heartbeat_timeout)
        heartbeat._start_heartbeat_observation()
        heartbeat._set_timestamp()
        result = None
        try:
            result = fn(*args, heartbeat=heartbeat, **kwargs)
        except HeartbeatTimeoutError:   # do not stop heartbeat observation -> task needs to be restarted
            return
        except RevokedError:    # stop heartbeat observation because task has been revoked
            pass
        except Exception as e:  # stop heartbeat observation and reraise exception
            heartbeat._stop_heartbeat_observation()
            raise e

        heartbeat._stop_heartbeat_observation()
        return result

    assert heartbeat_timeout >= 120, 'Minimal heartbeat_timeout is 120 seconds.'
    return inner


class RevokedError(Exception):
    pass


class HeartbeatTimeoutError(Exception):
    pass


class Heartbeat:

    CHECK_INTERVAL = timedelta(seconds=5)   # check timestamp just every 5 seconds -> otherwise redis can get heady load

    def __init__(self, huey: BaseHueyx, task: Task, heartbeat_timeout: int):
        self._huey = huey
        self.task = task
        self.heartbeat_timeout = heartbeat_timeout
        self.next_check = None

    @contextmanager
    def long_running_operation(self, delta: timedelta):
        self._set_timestamp(delta)
        yield
        self._set_timestamp()

    def __call__(self):
        if not self.next_check or self.next_check < timezone.now():
            self._check_timestamp()
            self.next_check = timezone.now() + self.CHECK_INTERVAL

    def _check_timestamp(self):
        """
        - Check if task has not been revoked -> RevokedError
        - Check if timestamp has not been expired -> HeartbeatTimeoutError
        Set new timestamp if checks are true
        """
        if self._huey.is_revoked(self.task):
            self._delete_timestamp()
            raise RevokedError()
        else:
            timestamp = self._get_timestamp()
            if not timestamp:
                raise HeartbeatTimeoutError()
            now = timezone.now()
            if timestamp + timedelta(seconds=self._huey.HEARTBEAT_UPDATE_INTERVAL) <= now:
                if timestamp + timedelta(seconds=self.heartbeat_timeout) <= now:
                    raise HeartbeatTimeoutError()
                self._set_timestamp()

    def _start_heartbeat_observation(self):
        """ Start heartbeat observation and save data to restart task if necessary. """
        args = self.task.args
        kwargs = self.task.kwargs
        if kwargs:
            kwargs.pop('task')
        task_settings = dict(on_complete=self.task.on_complete,
                             retries=self.task.retries, retry_delay=self.task.retry_delay, args=args, kwargs=kwargs)
        task_name = self.task.__module__ + '.' + self.task.name
        self._huey.put(self._huey.get_heartbeat_observation_key(self.task.id),
                       (task_name, task_settings, self.heartbeat_timeout))

    def _stop_heartbeat_observation(self):
        self._huey.get(self._huey.get_heartbeat_observation_key(self.task.id))

    def _set_timestamp(self, delta=timedelta()):
        self._huey.put(self._huey.get_heartbeat_timestamp_key(self.task.id), timezone.now() + delta)

    def _get_timestamp(self):
        return self._huey.get(self._huey.get_heartbeat_timestamp_key(self.task.id), peek=True)

    def _delete_timestamp(self):
        return self._huey.get(self._huey.get_heartbeat_timestamp_key(self.task.id), peek=False)



import json
import os
from importlib import import_module
from typing import Dict

from cached_property import cached_property
from django.conf import settings
from huey.storage import RedisStorage
from redis import ConnectionPool

from .redis_huey import RedisHuey


class HueyxException(Exception):
    pass


class SingleConfigReader:
    default_hueyx_path = 'hueyx.redis_huey'

    def __init__(self, name: str, config: Dict):
        self.name = name
        self.config = config
        self._is_prometheus_initialized = False

    @property
    def consumer_options(self):
        if 'prometheus_metrics_enabled' in self.config:
            self.config.pop('prometheus_metrics_enabled')
        if 'consumer' not in self.config:
            raise HueyxException('No consumer configured.')
        return self.config['consumer']

    @cached_property
    def connection_pool(self) -> ConnectionPool:
        if 'connection' not in self.config:
            raise HueyxException('No connection configured.')

        connection = self.config['connection']
        if 'connection_pool' in connection:
            return connection['connection_pool']
        else:
            return ConnectionPool(**connection)

    @property
    def huey_options(self):
        config = self.config.copy()
        config.pop('connection', {})
        config.pop('consumer', {})
        config.pop('backend_class', {})
        config.pop('name', {})
        return config

    @cached_property
    def huey_instance(self):
        huey_config = self.huey_options
        backend_path = huey_config.pop('huey_class', 'huey.RedisHuey')

        try:
            original_huey_module_path, class_name = backend_path.rsplit('.', 1)
            module = import_module('hueyx.redis_huey')
            backend_cls = getattr(module, class_name)
            huey = backend_cls(self.name, **huey_config, global_registry=False, connection_pool=self.connection_pool)
            if isinstance(huey.storage, RedisStorage) and self.is_signals_enabled:
                self._connect_signals_to_redis(huey)
            return huey
        except Exception:
            msg = f'Not supported Huey class: {backend_path}'
            raise HueyxException(msg)

    @cached_property
    def redis(self):
        return self.huey_instance.storage.conn

    @property
    def environment(self):
        if not 'environment' in settings.HUEYX_SIGNALS:
            return 'default'

        return settings.HUEYX_SIGNALS['environment']

    @property
    def is_signals_enabled(self):
        if not hasattr(settings, 'HUEYX_SIGNALS'):
            return False
        if not 'enabled' in settings.HUEYX_SIGNALS:
            return False
        return settings.HUEYX_SIGNALS['enabled']

    def _connect_signals_to_redis(self, huey: RedisHuey):
        huey._signal.connect(self._on_signal_received)

    def _on_signal_received(self, signal, task, exc=None):
        queue = self.huey_instance.name
        pid = os.getpid()
        data = {
            'environment': self.environment,
            'queue': queue,
            'pid': pid,
            'signal': signal,
            'task': task.name
        }
        self.redis.publish('hueyx.huey2.signaling', json.dumps(data))


class DjangoSettingsReader:
    def __init__(self):
        self.configurations: Dict[str, SingleConfigReader] = {}

    def interpret_settings(self):
        if not hasattr(settings, 'HUEYX'):
            raise HueyxException('No HUEYX config found in settings')

        for name, values in settings.HUEYX.items():
            reader = SingleConfigReader(name, values)
            self.configurations[name] = reader



from datetime import timedelta
from unittest.mock import MagicMock, patch

from django.test import TestCase
from django.utils import timezone
from huey.api import Task
from huey.constants import EmptyData

from hueyx.redis_huey import RedisHuey, Heartbeat, HeartbeatTimeoutError, RevokedError, _wrap_heartbeat


class RedisHueyMock(RedisHuey):

    def get_storage(self, *args, **kwargs):
        return MagicMock()

    def is_revoked(self, task, dt=None, peek=True):
        return False


class HeartbeatTest(TestCase):

    def example_func(self):
        return 0

    def setUp(self, *args):
        self.huey = RedisHueyMock()
        self.task = Task(retries=5, retry_delay=3)
        self.timeout = 120
        self.heartbeat = Heartbeat(self.huey, self.task, self.timeout)
        self.redis: MagicMock = self.huey.storage
        self.called = False
        self.call_cnt = 0

    def test_start_heartbeat_observation(self):
        self.heartbeat._start_heartbeat_observation()
        self.assertEqual(self.redis.method_calls[0][0], 'put_data')
        self.assertEqual(self.redis.method_calls[0][1][0], f'hb:{self.task.id}')

    def test_stop_heartbeat_observation(self):
        def pop_data(*args, **kwargs):
            self.assertEqual(args[0], f'hb:{self.task.id}')
            self.called = True
            return EmptyData

        self.redis.pop_data = pop_data
        self.heartbeat._stop_heartbeat_observation()
        self.assertTrue(self.called)

    def test_set_timestamp(self):
        self.heartbeat._set_timestamp()
        self.assertEqual(self.redis.method_calls[0][0], 'put_data')
        self.assertEqual(self.redis.method_calls[0][1][0], f'hbts:{self.task.id}')

    def test_get_timestamp(self):
        def peek_data(*args, **kwargs):
            self.assertEqual(args[0], f'hbts:{self.task.id}')
            self.called = True
            return EmptyData

        self.redis.peek_data = peek_data
        self.heartbeat._get_timestamp()
        self.assertTrue(self.called)

    def test_delete_timestamp(self):
        def pop_data(*args, **kwargs):
            self.assertEqual(args[0], f'hbts:{self.task.id}')
            self.called = True
            return EmptyData

        self.redis.pop_data = pop_data
        self.heartbeat._delete_timestamp()
        self.assertTrue(self.called)

    def test_heartbeat_no_new_timestamp(self):
        def get_timestamp():
            self.called = True
            return timezone.now()

        self.heartbeat._get_timestamp = get_timestamp
        self.heartbeat()
        self.assertTrue(self.called)

    def test_heartbeat_new_timestamp(self):
        def get_timestamp():
            return timezone.now() - timedelta(seconds=self.huey.HEARTBEAT_UPDATE_INTERVAL)

        def set_timestamp():
            self.called = True

        self.heartbeat._get_timestamp = get_timestamp
        self.heartbeat._set_timestamp = set_timestamp
        self.heartbeat()
        self.assertTrue(self.called)

    def test_heartbeat_no_timestamp(self):
        def get_timestamp():
            return None

        def set_timestamp():
            self.called = True

        self.heartbeat._get_timestamp = get_timestamp
        self.heartbeat._set_timestamp = set_timestamp
        with self.assertRaises(HeartbeatTimeoutError):
            self.heartbeat()
        self.assertFalse(self.called)

    def test_heartbeat_expired_timestamp(self):
        def get_timestamp():
            return timezone.now() - timedelta(seconds=self.timeout)

        def set_timestamp():
            self.called = True

        self.heartbeat._get_timestamp = get_timestamp
        self.heartbeat._set_timestamp = set_timestamp
        with self.assertRaises(HeartbeatTimeoutError):
            self.heartbeat()
        self.assertFalse(self.called)

    def test_heartbeat_revoked(self):
        def delete_timestamp():
            self.called = True

        self.huey.is_revoked = MagicMock(return_value=True)
        self.heartbeat._delete_timestamp = delete_timestamp
        with self.assertRaises(RevokedError):
            self.heartbeat()
        self.assertTrue(self.called)

    def test_caching(self):
        def get_timestamp():
            self.call_cnt += 1
            return timezone.now() - timedelta(seconds=self.huey.HEARTBEAT_UPDATE_INTERVAL)

        self.heartbeat._get_timestamp = get_timestamp
        self.heartbeat()
        self.heartbeat()
        self.assertEqual(self.call_cnt, 1)

    def test_no_caching(self):
        def get_timestamp():
            self.call_cnt += 1
            return timezone.now() - timedelta(seconds=self.huey.HEARTBEAT_UPDATE_INTERVAL)

        self.heartbeat._get_timestamp = get_timestamp
        self.heartbeat.CHECK_INTERVAL = timedelta()
        self.heartbeat()
        self.heartbeat()
        self.assertEqual(self.call_cnt, 2)


class HeartbeatMock(MagicMock):

    def _start_heartbeat_observation(self):
        self.calls = ['start']

    def _set_timestamp(self):
        self.calls.append('timestamp')

    def _stop_heartbeat_observation(self):
        self.calls.append('stop')


@patch('hueyx.redis_huey.Heartbeat', new_callable=HeartbeatMock)
class HeartbeatWrapperTest(TestCase):
    def setUp(self):
        self.huey = RedisHueyMock()
        self.task = Task(retries=5, retry_delay=3)
        self.timeout = 120
        self.heartbeat = None

    def test_task_execution(self, *args):
        def task(heartbeat):
            self.heartbeat = heartbeat
            return 'finish'

        result = _wrap_heartbeat(task, self.huey, self.timeout)(task=self.task)
        self.assertEqual(result, 'finish')
        self.assertEqual(self.heartbeat.calls, ['start', 'timestamp', 'stop'])

    def test_timeout(self, *args):
        def task(heartbeat):
            self.heartbeat = heartbeat
            raise HeartbeatTimeoutError()

        result = _wrap_heartbeat(task, self.huey, self.timeout)(task=self.task)
        self.assertEqual(result, None)
        self.assertEqual(self.heartbeat.calls, ['start', 'timestamp'])

    def test_revoke(self, *args):
        def task(heartbeat):
            self.heartbeat = heartbeat
            raise RevokedError()

        result = _wrap_heartbeat(task, self.huey, self.timeout)(task=self.task)
        self.assertEqual(result, None)
        self.assertEqual(self.heartbeat.calls, ['start', 'timestamp', 'stop'])

    def test_exception(self, *args):
        def task(heartbeat):
            self.heartbeat = heartbeat
            raise Exception()

        with self.assertRaises(Exception):
            _wrap_heartbeat(task, self.huey, self.timeout)(task=self.task)
        self.assertEqual(self.heartbeat.calls, ['start', 'timestamp', 'stop'])


class RedisMock(MagicMock):

    @property
    def conn(self, *args, **kwargs):
        return self

    # noinspection PyMethodMayBeStatic
    def hscan_iter(self, *args, **kwargs):
        return [[b'hb:task-id']]


class RedisHueyTest(TestCase):

    # @patch('huey.storage.RedisStorage', new_callable=RedisMock)
    def setUp(self, *args):
        self.huey = RedisHuey()
        self.huey.storage = RedisMock()
        self.called = False

    def test_no_dead_tasks(self):
        def get(key, *args, peek, **kwargs):
            assert peek
            self.called = True
            if key.startswith('hb:'):
                return 'name', 'settings', 60
            return timezone.now()

        self.huey.get = get
        tasks = self.huey.get_dead_tasks()
        self.assertEqual(len(tasks), 0)
        self.assertTrue(self.called)

    def test_dead_tasks_no_timestamp(self):
        def get(key, *args, peek, **kwargs):
            assert peek
            self.called = True
            if key.startswith('hb:'):
                return 'name', 'settings', 60

        self.huey.get = get
        tasks = self.huey.get_dead_tasks()
        self.assertEqual(len(tasks), 1)
        self.assertEqual(tasks[0], self.huey.DeadTask('task-id', 'name', 'settings'))
        self.assertTrue(self.called)

    def test_dead_tasks_timestamp_expired(self):
        def get(key, *args, peek, **kwargs):
            assert peek
            self.called = True
            if key.startswith('hb:'):
                return 'name', 'settings', 60
            return timezone.now() - timedelta(seconds=60)

        self.huey.get = get
        tasks = self.huey.get_dead_tasks()
        self.assertEqual(len(tasks), 1)
        self.assertEqual(tasks[0], self.huey.DeadTask('task-id', 'name', 'settings'))
        self.assertTrue(self.called)



from django.test import TestCase


HUEYX = {
    'queue1': {
        'connection': {
            'host': 'localhost',
            'port': 6379,
            'db': 98
        },
        'consumer': {
            'workers': 4,
            'worker_type': 'process',
        }
    },
    'queue2': {
        'connection': {
            'host': 'localhost',
            'port': 6379,
            'db': 99
        },
        'consumer': {
            'workers': 4,
            'worker_type': 'process',
        }
    },
}


class QueuesTest(TestCase):
    def test_settings_interpretation(self):
        with self.settings(HUEYX=HUEYX):
            from hueyx.queues import hueyx

            self.assertIsNotNone(hueyx('queue1'))
            self.assertIsNotNone(hueyx('queue2'))

    def test_huey_instance_caching(self):
        with self.settings(HUEYX=HUEYX):
            from hueyx.queues import hueyx

            self.assertEqual(hueyx('queue1'), hueyx('queue1'))




import datetime

import redis
import redis_lock
from huey.consumer import Consumer, Scheduler


class HueyxScheduler(Scheduler):
    def __init__(self, *args, **kwargs):
        self.multiple_scheduler_locking = kwargs.pop('multiple_scheduler_locking', False)
        super().__init__(*args, **kwargs)

    """
    Extend the usual Scheduler with the ability to prevent multiple periodic task execution due to multiple
    huey worker and finally multiple running Schedulers.
    This is done by locking the specific execution time pattern on redis.
    """

    def enqueue_periodic_tasks(self, now):
        self._logger.debug('Checking periodic tasks')
        if now is None:
            now = datetime.datetime.now()
        for task in self.huey.read_periodic(now):
            if self.check_and_set_for_multiple_execution(task, now):
                self.enqueue_periodic_task(task)
        return True

    def check_and_set_for_multiple_execution(self, task, now):
        """
        Checks if the task already has been scheduled from another huey worker and if not, marks the task as scheduled.
        :param task:
        :param now:
        :return: Bool if task can be scheduled.
        """

        if not self.multiple_scheduler_locking:
            return True

        conn: redis.ConnectionPool = self.huey.storage.conn
        full_name = f"huey.{self.huey.name}.{task.name}"
        lock_name = full_name + ".periodic_lock"
        pattern_name = full_name + '.time_pattern'
        with redis_lock.Lock(conn, lock_name, expire=60):
            if not self._can_execute(pattern_name, now):
                self._logger.info(
                    f'{full_name}: Do not schedule periodic task because this time pattern has already been scheduled.'
                )
                return False
            else:
                self._set_execution_time_pattern(pattern_name, now)
                self._logger.info(
                    f'{full_name}: Set time pattern for periodic task execution.'
                )
                return True

    def enqueue_periodic_task(self, task):
        self._logger.info('Scheduling periodic task %s.', task)
        self.huey.enqueue(task)

    def _create_time_pattern(self, now: datetime.datetime):
        """Standardized time pattern which the task is executed."""
        _, month, day, hour, minute, _, week_day, _, _ = now.timetuple()
        current_time_pattern = f"month{month}.day{day}.week_day{week_day}.hour{hour}.minute{minute}"
        return current_time_pattern

    def _can_execute(self, full_name: str, now: datetime.datetime):
        """Check if the current time pattern already has been executed."""
        current_time_pattern = self._create_time_pattern(now)

        last_time_pattern = self._get_last_execution_time_pattern(full_name)
        return last_time_pattern != current_time_pattern

    def _get_last_execution_time_pattern(self, full_name: str):
        conn = self.huey.storage.conn
        current_val = conn.get(full_name)
        if current_val is not None:
            current_val = current_val.decode("utf-8")
        return current_val

    def _set_execution_time_pattern(self, full_name: str, now):
        current_time_pattern = self._create_time_pattern(now)
        conn: redis.Redis = self.huey.storage.conn
        conn.set(full_name, current_time_pattern)


class HueyxConsumer(Consumer):
    def __init__(self, *args, **kwargs):
        self.multiple_scheduler_locking = kwargs.pop('multiple_scheduler_locking', False)
        super().__init__(*args, **kwargs)

    def _create_scheduler(self):
        self._logger.info('multiple_scheduler_locking: ' + str(self.multiple_scheduler_locking))
        return HueyxScheduler(
            huey=self.huey,
            interval=self.scheduler_interval,
            periodic=self.periodic,
            multiple_scheduler_locking=self.multiple_scheduler_locking)

    def run(self):
        self.huey.restart_dead_tasks()
        super().run()



import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='hueyx',
    version='1.0.3-dev2',
    packages=find_packages(),
    include_package_data=True,
    license='MIT License',  # example license
    description='Django huey extension which supports multiple huey queues.',
    long_description=README,
    long_description_content_type="text/markdown",
    url='https://github.com/APGSGA/hueyx',
    author='Severin Bühler',
    author_email='severin.buehler@apgsga.ch',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
    install_requires=[
        'cached-property',
        'huey>=2.3.0',
        'redis',
        'python-redis-lock',
    ],
)
